/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "factory.h"
#include "object.h"
#include "property.h"

#define REGISTER_BASIC_TYPE(Class,Type) \
    register_property_factory_type( #Class, #Type, sigc::ptr_fun(Class<Type>::create_object) ); \
    register_property_factory_type( #Class, "vector<" #Type ">", sigc::ptr_fun(Class<std::vector<Type> >::create_object) );

#define REGISTER_BASIC_TYPES(Class) \
    REGISTER_BASIC_TYPE(Class,float) \
    REGISTER_BASIC_TYPE(Class,double) \
    REGISTER_BASIC_TYPE(Class,long double) \
    REGISTER_BASIC_TYPE(Class,short int) \
    REGISTER_BASIC_TYPE(Class,unsigned short int) \
    REGISTER_BASIC_TYPE(Class,int) \
    REGISTER_BASIC_TYPE(Class,unsigned int) \
    REGISTER_BASIC_TYPE(Class,long int) \
    REGISTER_BASIC_TYPE(Class,unsigned long int) \
    REGISTER_BASIC_TYPE(Class,char) \
    REGISTER_BASIC_TYPE(Class,unsigned char) \
    REGISTER_BASIC_TYPE(Class,bool) \
    REGISTER_BASIC_TYPE(Class,std::string) \
    REGISTER_BASIC_TYPE(Class,Glib::ustring)

namespace Idioskopos
{
  typedef std::map<Glib::ustring,sigc::slot<Object::pointer> > CreateMap;
  CreateMap property_create_map;
  CreateMap introspectable_create_map;
  bool is_xml_property_factory_initialized = false;

  void xml_property_factory_init() {
    if ( is_xml_property_factory_initialized )
      return;

    REGISTER_BASIC_TYPES(ReadOnlyProperty);
    REGISTER_BASIC_TYPES(WriteOnlyProperty);
    REGISTER_BASIC_TYPES(Property);

    is_xml_property_factory_initialized = true;
  }

  const Glib::ustring & xml_dtd( bool as_inline, bool with_header )
  {
    static Glib::ustring ret;
    static const Glib::ustring dtd =
" \
<!ELEMENT introspectable ((introspectable|property)*) > \
<!ATTLIST introspectable type     CDATA  #REQUIRED    > \
<!ATTLIST introspectable name     CDATA  #REQUIRED    > \
\
<!ELEMENT property (value|array|vector)               > \
<!ATTLIST property       type     CDATA  #REQUIRED    > \
<!ATTLIST property       name     CDATA  #REQUIRED    > \
\
<!ELEMENT value (#CDATA)                              > \
<!ATTLIST value          type     CDATA  #REQUIRED    > \
\
<!ELEMENT array (value*)                              > \
<!ATTLIST array          type     CDATA  #REQUIRED    > \
<!ATTLIST array          size     CDATA  #REQUIRED    > \
\
<!ELEMENT vector (value*)                             > \
<!ATTLIST vector         type     CDATA  #REQUIRED    > \
<!ATTLIST vector         size     CDATA  #REQUIRED    > \
";
    if ( as_inline || with_header ) {
      ret = dtd;
      if ( with_header )
        ret = "<?xml version=\"1.0\"?>\n" + ret;
      if ( as_inline ) {
        ret = "<!DOCTYPE introspectable [\n" + ret + "]>\n";
      }
      return ret;
    }
    return dtd;
  }

  void register_property_factory_type( const Glib::ustring& class_name,
                                       const Glib::ustring& value_type,
                                       sigc::slot<Object::pointer> slot )
  {
    Glib::ustring index;
    index = class_name + "-" + value_type;
    property_create_map[index] = slot;
  }

  void unregister_property_factory_type( const Glib::ustring& class_name,
                                       const Glib::ustring& value_type )
  {
    Glib::ustring index;
    index = class_name + "-" + value_type;
    property_create_map.erase(index);
  }

  void register_introspectable_factory_type( const Glib::ustring& class_name,
                                             sigc::slot<Object::pointer> slot )
  {
    introspectable_create_map[class_name] = slot;
  }

  void unregister_introspectable_factory_type( const Glib::ustring& class_name )
  {
    introspectable_create_map.erase(class_name);
  }

  Object::pointer xml_factory_create( const Glib::ustring & xml_initialization )
  {
    xmlpp::DomParser parser;
    parser.parse_memory(xml_initialization);
    if (parser)
      return xml_factory_create( parser.get_document()->get_root_node() );
    return Object::pointer();
  }

  Object::pointer xml_factory_create( const xmlpp::Node * n )
  {
    const xmlpp::Element *element;

    element = dynamic_cast<const xmlpp::Element*>(n);
    if ( !element ) return Object::pointer();

    if ( element->get_name() == "property" ) return xml_property_factory_create(n);

    if ( element->get_name() == "introspectable" ) return xml_introspectable_factory_create(n);

    return Object::pointer();
  }

  Object::pointer xml_property_factory_create( const xmlpp::Node * n )
  {
    const xmlpp::Element *element, *value_element;
    const xmlpp::Attribute *typeattr, *valtypeattr;
    Glib::ustring class_name, value_type, index;
    xmlpp::NodeSet values;
    Object::pointer created;

    if ( ! is_xml_property_factory_initialized )
      xml_property_factory_init();

    element = dynamic_cast<const xmlpp::Element*>(n);
    if ( !element ) return Object::pointer();

    if ( element->get_name() != "property" ) return Object::pointer();

    typeattr = element->get_attribute("type");
    if ( !typeattr ) return Object::pointer();

    class_name = typeattr->get_value();

    values = element->find("value");
    if ( values.size() != 0 ) {
      value_element = dynamic_cast<const xmlpp::Element*>(values.front());
      if ( !value_element ) return Object::pointer();
      valtypeattr = value_element->get_attribute("type");
      if ( !valtypeattr ) return Object::pointer();
      value_type = valtypeattr->get_value();

      index = class_name + "-" + value_type;

      CreateMap::iterator iter;
      iter = property_create_map.find(index);

      if ( iter == property_create_map.end() )
        return Object::pointer();

      sigc::slot<Object::pointer> creator = iter->second;

      Object::pointer obj = creator();

      obj->set_xml( n, true );

      return obj;
    }

    values = element->find("vector");
    if ( values.size() == 0 )
      values = element->find("array");
    if ( values.size() != 0 ) {
      value_element = dynamic_cast<const xmlpp::Element*>(values.front());
      if ( !value_element ) return Object::pointer();
      valtypeattr = value_element->get_attribute("type");
      if ( !valtypeattr ) return Object::pointer();
      value_type = valtypeattr->get_value();

      index = class_name + "-vector<" + value_type + ">";

      CreateMap::iterator iter;
      iter = property_create_map.find(index);

      if ( iter == property_create_map.end() )
        return Object::pointer();

      Object::pointer obj = iter->second();

      obj->set_xml( n, true );

      return obj;
    }

    return Object::pointer();
  }

  Object::pointer xml_introspectable_factory_create( const xmlpp::Node * n )
  {
    const xmlpp::Element *element;
    const xmlpp::Attribute *typeattr;
    Glib::ustring class_name;
    Object::pointer created;

    element = dynamic_cast<const xmlpp::Element*>(n);
    if ( !element ) return Object::pointer();

    if ( element->get_name() != "introspectable" ) return Object::pointer();

    typeattr = element->get_attribute("type");
    if ( !typeattr ) return Object::pointer();

    class_name = typeattr->get_value();

    CreateMap::iterator iter;

    iter = introspectable_create_map.find(class_name);

    if ( iter == introspectable_create_map.end() )
      return Object::pointer();

    sigc::slot<Object::pointer> creator = iter->second;

    Object::pointer obj = creator();

    obj->set_xml( n, true, true );

    return obj;
  }

}


