/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "introspectable.h"
#include "factory.h"

#include <iostream>
#include <sstream>

namespace Idioskopos
  {

  IDIOSKOPOS_REGISTRAR(Introspectable);

  Introspectable::Introspectable(const Glib::ustring& name, sigc::slot<void> slot):
      Object(name, slot),
      m_separator(".")
  {}

  Introspectable::Introspectable(const Glib::ustring& name,
                                 Introspectable& container,
                                 sigc::slot<void> slot):
      Object(name, container, slot),
      m_separator(".")
  {}

  Introspectable::pointer Introspectable::create( const Glib::ustring & name, sigc::slot<void> slot )
  {
    return pointer( new Introspectable(name, slot) );
  }

  Introspectable::pointer Introspectable::create( const Glib::ustring & name,
                                                  Introspectable& container,
                                                  sigc::slot<void> slot )
  {
    return pointer( new Introspectable(name, container, slot) );
  }

  Introspectable::~Introspectable()
  {}

  Object* Introspectable::get_property(const Glib::ustring& name)
  {
    PropertyMap::iterator i;
    i = m_property_name_map.find(name);
    if (i != m_property_name_map.end())
      return i->second;

    return NULL;
  }

  bool Introspectable::has_property(const Glib::ustring& name)
  {
    Object* sp = get_property(name);
    return (sp != NULL);
  }

  bool Introspectable::add_property( Object & property )
  {
    if (this->has_property(property.name()))
      return false;
    m_properties.insert( &property );
    m_property_name_map[property.name()] = &property;
    m_property_value_changed_connection[&property] =
        property.signal_value_changed().connect(sigc::bind(sigc::mem_fun(*this, &Introspectable::on_property_value_changed), &property));
    m_property_name_changed_connection[&property] =
        property.signal_name_changed().connect(sigc::bind(sigc::mem_fun(*this, &Introspectable::on_property_name_changed), &property));
    m_signal_property_added.emit(&property);
    return true;
  }

  bool Introspectable::add_property( Object::pointer property )
  {
    bool added;
    if ( ! property )
      return false;
    added = this->add_property( *property );
    if ( added )
      m_dynamic_property_map[ property.get() ] = property;
    return added;
  }

  void Introspectable::remove_property(const Glib::ustring& name)
  {
    Object * property = get_property(name);

    if (property == NULL)
      return;

    m_property_value_changed_connection[property].disconnect();
    m_property_name_changed_connection[property].disconnect();
    m_properties.erase(property);
    m_property_value_changed_connection.erase(property);
    m_property_name_changed_connection.erase(property);
    m_property_name_map.erase(name);
    m_signal_property_removed.emit(property);
    m_dynamic_property_map.erase(property);
  }

  sigc::signal<void, Object*> Introspectable::signal_property_value_changed()
  {
    return m_signal_property_value_changed;
  }

  sigc::signal<void, Object*, const Glib::ustring&> Introspectable::signal_property_name_changed()
  {
    return m_signal_property_name_changed;
  }

  sigc::signal< void, Object * > Introspectable::signal_property_added( )
  {
    return m_signal_property_added;
  }

  sigc::signal< void, Object * > Introspectable::signal_property_removed( )
  {
    return m_signal_property_removed;
  }

  Glib::ustring Introspectable::xml( int indent )
  {
    std::set<Object*>::iterator iter;
    Glib::ustring s;
    int i;
    for (i=0; i < indent; i++) s += "  ";
    s += "<introspectable type=\"" + this->class_name()  + "\" name=\"" + this->name() + "\">";
    if ( indent >= 0 ) s += "\n";
    for (iter = m_properties.begin(); iter != m_properties.end(); iter++)
      s += (*iter)->xml((indent>=0)?indent+1:indent);
    for (i=0; i < indent; i++) s += "  ";
    s += "</introspectable>";
    if ( indent >= 0 ) s += "\n";
    return s;
  }

  void Introspectable::set_xml(const Glib::ustring& s, bool change_names, bool create_missing)
  {
    Object::set_xml(s, change_names, create_missing);
  }

  void Introspectable::set_xml(const xmlpp::Node* n, bool change_names, bool create_missing)
  {
    xmlpp::Attribute* attribute;
    xmlpp::NodeSet nodes;
    xmlpp::NodeSet::iterator iter;
    const xmlpp::Element *elem, *elem2;
    Object* obj;

    elem = dynamic_cast<const xmlpp::Element*>(n);
    if ( !elem ) return;

    if ( change_names )
    {
      attribute = elem->get_attribute("name");
      if ( attribute )
        this->set_name( attribute->get_value() );
    }

    nodes = elem->find("property");

    for ( iter=nodes.begin(); iter != nodes.end(); iter++ )
    {
      elem2 = dynamic_cast<const xmlpp::Element*>(*iter);
      if ( !elem2 ) continue;
      attribute = elem2->get_attribute("name");
      if ( !attribute ) continue;
      obj = this->get_property( attribute->get_value() );
      if ( obj )
         obj->set_xml( elem2, change_names, create_missing );
      else if ( create_missing ) {
        Object::pointer created;
        created = xml_property_factory_create( *iter );
        if ( created )
          this->add_property(created);
      }
    }

    nodes = elem->find("introspectable");

    for ( iter=nodes.begin(); iter != nodes.end(); iter++ )
    {
      elem2 = dynamic_cast<const xmlpp::Element*>(*iter);
      if ( !elem2 ) continue;
      attribute = elem2->get_attribute("name");
      if ( !attribute ) continue;
      obj = this->get_property( attribute->get_value() );
      if ( obj )
        obj->set_xml( elem2, change_names, create_missing );
      else if ( create_missing ) {
        Object::pointer created;
        created = xml_introspectable_factory_create( *iter );
        if ( created )
          this->add_property(created);
      }
    }

    return;
  }

  size_t Introspectable::num_properties() const
  {
    return m_property_name_map.size();
  }

  Object::iterator Introspectable::properties_begin()
  {
    return Object::iterator(this, m_properties.begin());
  }

  Object::iterator Introspectable::properties_end()
  {
    return Object::iterator(this, m_properties.end());
  }

  Object::iterator Introspectable::next_property(std::set<Object*>::iterator i)
  {
    if ( i != m_properties.end() )
      return Object::iterator(this, ++i);
    return Object::iterator();
  }

  Object::iterator Introspectable::previous_property(std::set<Object*>::iterator i)
  {
    if ( i != m_properties.begin() )
      return Object::iterator(this, --i);
    return Object::iterator();
  }

  Object& Introspectable::operator[](const Glib::ustring& name) throw (error::bad_property_name)
  {
    Object* pb = get_property(name);
    if (pb == NULL)
      throw error::bad_property_name();
    return *pb;
  }

  void Introspectable::on_property_value_changed(Object* object)
  {
    m_signal_property_value_changed.emit(object);
  }

  void Introspectable::on_property_name_changed(const Glib::ustring& oldname, Object* object)
  {
    m_property_name_map.erase(oldname);
    m_property_name_map[object->name()] = object;
    m_signal_property_name_changed.emit(object, oldname);
  }

}

