/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "object.h"

#include <idioskopos/introspectable.h>

namespace Idioskopos
{

  Object::iterator::iterator() : m_type( NONE ), m_container( NULL ), m_index( 0 )
  { }

  Object::iterator::iterator( Object* container, std::set
                                <Object*>::iterator iter ) :
        m_type( SET_ITERATOR ),
        m_container( container ),
        m_set_iterator( iter ),
        m_index( 0 )
  { }

  Object::iterator::iterator( Object* container, Object::pointer containee, size_t index ) :
      m_type( INDEX ),
      m_container( container ),
      m_containee( containee ),
      m_index( index )
  { }

  Object::iterator& Object::iterator::operator++()
  {
    if ( m_container )
      switch ( m_type ) {
        case SET_ITERATOR:
          *this = m_container->next_property( m_set_iterator );
          break;
        case INDEX:
          *this = m_container->next_property( m_index );
          break;
        case NONE:
          break;
      }
    return *this;
  }

  Object::iterator& Object::iterator::operator--()
  {
    if ( m_container )
      switch ( m_type ) {
        case SET_ITERATOR:
          *this = m_container->previous_property( m_set_iterator );
          break;
        case INDEX:
          *this = m_container->previous_property( m_index );
          break;
        case NONE:
          break;
      }
    return *this;
  }

  Object::iterator Object::iterator::operator++( int )
  {
    iterator i = *this;
    if ( m_container )
      switch ( m_type ) {
        case SET_ITERATOR:
          *this = m_container->next_property( m_set_iterator );
          break;
        case INDEX:
          *this = m_container->next_property( m_index );
          break;
        case NONE:
          break;
      }
    return i;
  }

  Object::iterator Object::iterator::operator--( int )
  {
    iterator i = *this;
    if ( m_container )
      switch ( m_type ) {
        case SET_ITERATOR:
          *this = m_container->previous_property( m_set_iterator );
          break;
        case INDEX:
          *this = m_container->previous_property( m_index );
          break;
        case NONE:
          break;
      }
    return i;
  }

  Object* Object::iterator::operator->()
  {
    switch ( m_type ) {
      case SET_ITERATOR:
        return * m_set_iterator;
      case INDEX:
        return m_containee.get();
      case NONE:
        break;
    }
    return NULL;
  }

  Object& Object::iterator::operator*()
  {
    switch ( m_type ) {
      case SET_ITERATOR:
        return **m_set_iterator;
      case INDEX:
        return *m_containee;
      case NONE:
        break;
    }
    throw;
  }

  bool Object::iterator::operator==( const iterator& other ) const
  {
    if ( m_type != other.m_type )
      return false;
    switch ( m_type ) {
      case NONE:
        return other.m_type == NONE;
      case SET_ITERATOR:
        return ( m_container == other.m_container && *m_set_iterator == *( other.m_set_iterator ) );
      case INDEX:
        return ( m_container == other.m_container && m_index == other.m_index );
    }
    return false;
  }

  bool Object::iterator::operator!=( const iterator& other ) const
  {
    return ! ( *this == other );
  }

  Object::Object( const Glib::ustring& name, sigc::slot<void> slot ) :
      m_name( name )
  {
    if ( slot )
      m_signal_value_changed.connect( slot );
  }

  Object::Object( const Glib::ustring& name, Introspectable& container, sigc::slot<void> slot ) :
      m_name( name )
  {
    container.add_property( *this );
    if ( slot )
      m_signal_value_changed.connect( slot );
  }

  Object::~Object()
  {}

  const Glib::ustring& Object::name()
  {
    return m_name;
  }

  void Object::set_name( const Glib::ustring& n )
  {
    Glib::ustring old_name = m_name;
    m_name = n;
    m_signal_name_changed.emit( old_name );
  }

  Glib::ustring Object::value()
  {
    return Glib::ustring();
  }

  void Object::set_value( const Glib::ustring& v )
  {}

  void Object::set_xml( const Glib::ustring& s, bool change_names, bool create_missing )
  {
    xmlpp::DomParser parser;
    parser.parse_memory( s );
    if ( parser ) {
      this->set_xml( parser.get_document() ->get_root_node(), change_names, create_missing );
    }
  }

  void Object::set_xml( const xmlpp::Node* n, bool change_names, bool create_missing )
  {}

  size_t Object::num_properties() const
  {
    return 0;
  }

  Object::iterator Object::properties_begin()
  {
    return properties_end();
  }

  Object::iterator Object::properties_end()
  {
    return iterator();
  }

  Object& Object::operator[] ( const Glib::ustring& name ) throw( error::bad_property_name )
  {
    throw error::bad_property_name();
  }

  sigc::signal< void, Object * > Object::signal_property_added( )
  {
    throw;
  }

  sigc::signal< void, Object * > Object::signal_property_removed( )
  {
    throw;
  }

  sigc::signal<void, const Glib::ustring&> Object::signal_name_changed()
  {
    return m_signal_name_changed;
  }

  sigc::signal<void>& Object::signal_value_changed()
  {
    return m_signal_value_changed;
  }

  Object::iterator Object::next_property( std::set<Object*>::iterator i )
  {
    return iterator();
  }

  Object::iterator Object::previous_property( std::set<Object*>::iterator i )
  {
    return iterator();
  }

  Object::iterator Object::next_property( size_t i )
  {
    return iterator();
  }

  Object::iterator Object::previous_property( size_t i )
  {
    return iterator();
  }

}

