/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/

#ifndef IDIOSKOPOSOBJECT_H
#define IDIOSKOPOSOBJECT_H

#include <string>
#include <set>
#include <iostream>
#include <sstream>
#include <iterator>
#include <sigc++/sigc++.h>
#include <libxml++/libxml++.h>
#include <typeinfo>

#include <idioskopos/pointer.h>
#include <idioskopos/error.h>

#define IDIOSKOPOS_OBJECT(idioskopos_name)           \
  virtual const Glib::ustring& class_name() const {  \
    static Glib::ustring s = (#idioskopos_name);     \
    return (s);                                      \
  }

/**
 * \mainpage Idioskopos User's Guide
 *
 * This is the user's guide to the Idioskopos library.
 *
 * \section toc Contents
 *
 * -# \ref quick_start
 * -# Properties
 *   -# \ref guide_property
 *   -# \ref guide_reference_property
 *   -# \ref guide_virtual_property
 * -# Introspection
 * -# \ref guide_introspectable
 */

/**
 * \page quick_start Quick Start Guide to Idioskopos
 *
 * For now, this example in simple.cpp will have to suffice as a guide.
 *
 * \include simple.cpp
 */

namespace Idioskopos
  {

  class Introspectable;

  /**
   * Serves as a common interface to all properties.
   *
   * Namely, it provides an top-level interface for signals and RTTI information.
   * Since Property is a template class that inherits from Object, this
   * serves as a common ancestor.
   *
   * It is useful for generic containers of properties, such as that used
   * by Introspectable.
   *
   * @author Rick L Vinyard Jr
   */
  class Object: public virtual sigc::trackable
    {
    public:
      typedef IdioskoposPointer<Object> pointer;
      typedef std::set<Object*> Objects;

      /** An STL type iterator to iterate through the children of an introspectable object */
      class iterator {
        public:
          iterator();
          iterator( Object* container, std::set<Object*>::iterator iter );
          iterator( Object* container, Object::pointer containee, size_t index );

          iterator& operator++();
          iterator& operator--();
          iterator operator++(int);
          iterator operator--(int);
          Object* operator->();
          Object& operator*();
          bool operator==(const iterator& other) const;
          bool operator!=(const iterator& other) const;

        protected:

          typedef enum TYPE {
            NONE,
            SET_ITERATOR,
            INDEX,
          } TYPE;

          TYPE m_type;
          Object* m_container;
          std::set<Object*>::iterator m_set_iterator;
          Object::pointer m_containee;
          size_t m_index;
      };

      IDIOSKOPOS_OBJECT(Object);

      Object(const Glib::ustring& name,
             sigc::slot<void> slot=sigc::slot<void>() );

      /**
       * Constructor that adds this property to the introspectable container
       * and will identify this property by the given name.
       */
      Object(const Glib::ustring& name,
             Introspectable& container,
             sigc::slot<void> slot=sigc::slot<void>());

      virtual ~Object();

      const Glib::ustring& name();

      void set_name(const Glib::ustring& n);

      /**
       * Returns the property added signal if this is an introspectable object,
       * otherwise throws an exception.
       */
      virtual sigc::signal<void, Object*> signal_property_added();

      /**
       * Returns the property removed signal if this is an introspectable object,
       * otherwise throws an exception.
       */
      virtual sigc::signal<void, Object*> signal_property_removed();

      sigc::signal<void,const Glib::ustring&> signal_name_changed();

      /**
       * Signal emitted when the property is changed.
       */
      sigc::signal<void>& signal_value_changed();

      /**
       * Pure virtual method implemented by children to return actual RTTI information from child.
       */
      virtual const std::type_info& type() const = 0;

      /**
       * Pure virtual method implemented by children to return an XML string containing
       * both type and value information.
       */
      virtual Glib::ustring xml(int indent=-1) = 0;

      virtual void set_xml(const Glib::ustring& s, bool change_names=false, bool create_missing=true);

      virtual void set_xml(const xmlpp::Node* n, bool change_names=false, bool create_missing=true);

      /**
       * Virtual method implemented by children to return a string representation of
       * the contained value.
       */
      virtual Glib::ustring value();

      virtual void set_value( const Glib::ustring& v );

      template <typename T>
          void set_value( const T& t )
      {
        std::ostringstream sout;
        Glib::ustring s;
        sout << t;
        s = sout.str();
        this->set_value(s);
      }

      /**
       * If this is a property container, this should be reimplemented
       * to provide a true count of the number of child properties this
       * container has.
       */
      virtual size_t num_properties() const;

      virtual iterator properties_begin();

      virtual iterator properties_end();

      /**
       * This method returns an Object reference and throws exceptions when a bad
       * name is encountered. The alternative would be to return an Object* and
       * return NULL for an unfound name. However, by returning a reference,
       * it is possible to index into objects using array notation, as in
       *      \em myobject["child"]["property"]
       */
      virtual Object& operator[](const Glib::ustring& name) throw(error::bad_property_name);

    protected:
      Glib::ustring m_name;

      sigc::signal<void, const Glib::ustring&> m_signal_name_changed;

      /**
       * Signal emitted when underlying data is changed.
       */
      sigc::signal<void> m_signal_value_changed;

      virtual iterator next_property(std::set<Object*>::iterator i);

      virtual iterator previous_property(std::set<Object*>::iterator i);

      virtual iterator next_property(size_t i);

      virtual iterator previous_property(size_t i);

    };

}

#include <idioskopos/factory.h>

#endif
