/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/

#ifndef IDIOSKOPOSPROPERTY_H
#define IDIOSKOPOSPROPERTY_H

#include <idioskopos/propertybase.h>
#include <idioskopos/array.h>

namespace Idioskopos {

  /**
   * Read only property
   */
  template <typename T>
  class ReadOnlyProperty: public ReadOnlyPropertyBase<T> {
    public:
      typedef IdioskoposPointer<ReadOnlyProperty> pointer;
      typedef T value_type;

      IDIOSKOPOS_OBJECT(ReadOnlyProperty);

      /**
       * Default constructor that doesn't do much.
       */
      ReadOnlyProperty ( const Glib::ustring& name,
                         sigc::slot<void> slot=sigc::slot<void>()):
          Object(name, slot),
          ReadOnlyPropertyBase<T>(name)
      { }

      /**
       * Initializes the contained value to the parameter's value.
       */
          ReadOnlyProperty ( const Glib::ustring& name,
                             const T& val,
                             sigc::slot<void> slot=sigc::slot<void>() ) :
          Object(name, slot),
          ReadOnlyPropertyBase<T>(name),
          m_value( val )
      { }

      /**
       * Makes this a property of the introspectable container accessible through
       * the provided name.
       */
      ReadOnlyProperty( const Glib::ustring& name,
                        Introspectable& container,
                        sigc::slot<void> slot=sigc::slot<void>() ) :
          Object(name, container, slot),
          ReadOnlyPropertyBase<T>( name, container )
      { }

      /**
       * Makes this a property of the introspectable container accessible through
       * the provided name, and initializes the contained value to the value of
       * the third parameter.
       */
      ReadOnlyProperty ( const Glib::ustring& name,
                         const T& val,
                         Introspectable& container,
                         sigc::slot<void> slot=sigc::slot<void>() ) :
          Object(name, container, slot),
          ReadOnlyPropertyBase<T>( name, container ),
          m_value( val )
      { }

      static Object::pointer create_object()
      {
        return Object::pointer( new ReadOnlyProperty<T>(Glib::ustring()) );
      }

      static pointer create(const Glib::ustring& name,
                            sigc::slot<void> slot=sigc::slot<void>())
      {
        return pointer( new ReadOnlyProperty<T>(name, slot) );
      }

      static pointer create( const Glib::ustring& name,
                             const T& val,
                             sigc::slot<void> slot=sigc::slot<void>() )
      {
        return pointer( new ReadOnlyProperty<T>(name, val, slot) );
      }

      static pointer create( const Glib::ustring& name,
                             Introspectable& container,
                             sigc::slot<void> slot=sigc::slot<void>() )
      {
        return pointer( new ReadOnlyProperty<T>(name, container, slot) );
      }

      static pointer create( const Glib::ustring& name,
                             const T& val,
                             Introspectable& container,
                             sigc::slot<void> slot=sigc::slot<void>())
      {
        return pointer( new ReadOnlyProperty<T>(name, val, container, slot) );
      }

      virtual ~ReadOnlyProperty() { }

      /** The canonical get() accessor method. */
      virtual const T& get() const { return m_value; }

      /** Should return a pointer to the contained object, or NULL if pointers are not allowed */
      virtual T* get_pointer() { return & m_value; }

    protected:
      T m_value;

  };

  /**
   * Write only property
   */
  template <typename T>
  class WriteOnlyProperty: public WriteOnlyPropertyBase<T> {
    public:
      typedef IdioskoposPointer<WriteOnlyProperty> pointer;
      typedef T value_type;

      IDIOSKOPOS_OBJECT(WriteOnlyProperty);

      /**
       * Default constructor that doesn't do much.
       */
      WriteOnlyProperty (const Glib::ustring& name, sigc::slot<void> slot=sigc::slot<void>()):
          Object(name, slot),
          WriteOnlyPropertyBase<T>(name) { }

      /**
       * Initializes the contained value to the parameter's value.
       */
          WriteOnlyProperty ( const Glib::ustring& name,
                              const T& val,
                              sigc::slot<void> slot=sigc::slot<void>() ) :
          Object(name, slot),
          WriteOnlyPropertyBase<T>(name),
          m_value( val ) { }

      /**
       * Makes this a property of the introspectable container accessible through
       * the provided name.
       */
      WriteOnlyProperty( const Glib::ustring& name,
                         Introspectable& container,
                         sigc::slot<void> slot=sigc::slot<void>() ) :
          Object(name, container, slot),
          WriteOnlyPropertyBase<T>( name, container ) { }

      /**
       * Makes this a property of the introspectable container accessible through
       * the provided name, and initializes the contained value to the value of
       * the third parameter.
       */
      WriteOnlyProperty ( const Glib::ustring& name,
                          const T& val,
                          Introspectable& container,
                          sigc::slot<void> slot=sigc::slot<void>() ) :
          Object(name, container, slot),
          WriteOnlyPropertyBase<T>( name, container ),
          m_value( val ) { }

      static Object::pointer create_object()
      {
        return Object::pointer( new WriteOnlyProperty<T>(Glib::ustring()) );
      }

      static pointer create(const Glib::ustring& name,
                            sigc::slot<void> slot=sigc::slot<void>())
      {
        return pointer( new WriteOnlyProperty<T>(name, slot) );
      }

      static pointer create( const Glib::ustring& name,
                             const T& val,
                             sigc::slot<void> slot=sigc::slot<void>() )
      {
        return pointer( new WriteOnlyProperty<T>(name, val, slot) );
      }

      static pointer create( const Glib::ustring& name,
                             Introspectable& container,
                             sigc::slot<void> slot=sigc::slot<void>() )
      {
        return pointer( new WriteOnlyProperty<T>(name, container, slot) );
      }

      static pointer create( const Glib::ustring& name,
                             const T& val,
                             Introspectable& container,
                             sigc::slot<void> slot=sigc::slot<void>() )
      {
        return pointer( new WriteOnlyProperty<T>(name, val, container, slot) );
      }

      virtual ~WriteOnlyProperty() { }

      /** The canonical set() accessor method that also emits the changed signal. */
      virtual void set( const T& val )
      {
        m_value = val;
        this->signal_value_changed().emit();
      }

    protected:
      T m_value;

      virtual T& get_value()
      {
        return m_value;
      }
  };

  /**
   * Property
   */
  template <typename T>
  class Property: public PropertyBase<T> {
    public:
      typedef IdioskoposPointer<Property> pointer;
      typedef T value_type;

      IDIOSKOPOS_OBJECT(Property);

      /**
       * Default constructor that doesn't do much.
       */
      Property (const Glib::ustring& name,
                sigc::slot<void> slot=sigc::slot<void>()):
          Object(name, slot),
          PropertyBase<T>(name),
          m_value()
      { }

      /**
       * Initializes the contained value to the parameter's value.
       */
      Property ( const Glib::ustring& name,
                 const T& val,
                 sigc::slot<void> slot=sigc::slot<void>() ) :
          Object(name, slot),
          PropertyBase<T>(name),
          m_value( val )
      { }

      /**
       * Makes this a property of the introspectable container accessible through
       * the provided name.
       */
      Property( const Glib::ustring& name,
                Introspectable& container,
                sigc::slot<void> slot=sigc::slot<void>() ) :
          Object( name, container, slot ),
          PropertyBase<T>(name, container),
          m_value()
      { }

      /**
       * Makes this a property of the introspectable container accessible through
       * the provided name, and initializes the contained value to the value of
       * the third parameter.
       */
      Property ( const Glib::ustring& name,
                 const T& val,
                 Introspectable& container,
                 sigc::slot<void> slot=sigc::slot<void>() ) :
          Object( name, container, slot ),
          PropertyBase<T>(name, container),
          m_value( val )
      { }

      static Object::pointer create_object()
      {
        return Object::pointer( new Property<T>(Glib::ustring()) );
      }

      static pointer create(const Glib::ustring& name,
                            sigc::slot<void> slot=sigc::slot<void>())
      {
        return pointer( new Property<T>(name, slot) );
      }

      static pointer create( const Glib::ustring& name,
                             const T& val,
                             sigc::slot<void> slot=sigc::slot<void>() )
      {
        return pointer( new Property<T>(name, val, slot) );
      }

      static pointer create( const Glib::ustring& name,
                             Introspectable& container,
                             sigc::slot<void> slot=sigc::slot<void>() )
      {
        return pointer( new Property<T>(name, container, slot) );
      }

      static pointer create( const Glib::ustring& name,
                             const T& val,
                             Introspectable& container,
                             sigc::slot<void> slot=sigc::slot<void>() )
      {
        return pointer( new Property<T>(name, val, container, slot) );
      }

      virtual ~Property() { }

      /** The canonical get() accessor method. */
      virtual const T& get() const { return m_value; }

      /** Should return a pointer to the contained object, or NULL if pointers are not allowed */
      virtual T* get_pointer()
      {
        return & m_value;
      }

      /** The canonical set() accessor method that also emits the changed signal. */
      virtual void set
        ( const T& val )
      {
        m_value = val;
        this->signal_value_changed().emit();
      }

      Property<T>& operator= (const T& val)
      {
        this->set(val);
        return *this;
      }

    protected:
      T m_value;

      virtual T& get_value()
      {
        return m_value;
      }
  };

}


#endif

