/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/

#ifndef IDIOSKOPOSPROPERTYBASE_H
#define IDIOSKOPOSPROPERTYBASE_H

#include <idioskopos/readonlypropertybase.h>
#include <idioskopos/writeonlypropertybase.h>

// TODO
// Need to make properties restrictable on the introspection point, with proxies to temporarily restrict or enhance permissions

namespace Idioskopos
{

  /**
   * Implements a simple property interface with predefined get/set methods.
   *
   * A property is intended to encapsulate a simple value that is
   * most typically modified through assignments, i.e. integers,
   * doubles, strings, et. al.
   *
   * @author Rick L Vinyard Jr
  */
  template <typename T>
  class PropertyBase: public ReadOnlyPropertyBase<T>, public WriteOnlyPropertyBase<T>
  {
    public:
      typedef IdioskoposPointer<PropertyBase> pointer;
      typedef T value_type;

      IDIOSKOPOS_OBJECT( PropertyBase );

      /**
       * Default constructor that doesn't do much.
       */
      PropertyBase ( const Glib::ustring& name,
                     sigc::slot<void> slot=sigc::slot<void>() ) :
          Object( name, slot ),
          ReadOnlyPropertyBase<T>( name ),
          WriteOnlyPropertyBase<T>( name )
      { }

      /**
       * Makes this a property of the introspectable container accessible through
       * the provided name.
       */
          PropertyBase( const Glib::ustring& name,
                        Introspectable& container,
                        sigc::slot<void> slot=sigc::slot<void>() ) :
          Object( name, container, slot ),
          ReadOnlyPropertyBase<T>( name, container ),
          WriteOnlyPropertyBase<T>( name, container )
      { }

      virtual ~PropertyBase()
      { }

      /**
       * Allows assignment to the property from the contained value type.
       */
      PropertyBase<T>& operator= ( const T& val )
      {
        this->set( val );
        return *this;
      }

      /**
       * Function call syntax
       *
       * Set the value of the property with property(value) notation.
       * @return The new value of the property
       */
      PropertyBase<T>& operator() ( const T& val )
      {
        this->set( val );
        return *this;
      }

      /**
       * Arithmetic assignment operator
       *
       * This method is implemented as a template for two reasons:
       *  - So that it will work with a variety of operands
       *  - So that a stub will not be generated if the operator is not defined for T
       *
       * T + X must be well defined
       */
      template <typename X>
      PropertyBase<T>& operator+=( X other )
      {
        this->set
        ( this->get
          () + other );
        return *this;
      }

      /**
       * Arithmetic assignment operator
       *
       * This method is implemented as a template for two reasons:
       *  - So that it will work with a variety of operands
       *  - So that a stub will not be generated if the operator is not defined for T
       *
       * T - X must be well defined
       */
      template <typename X>
      PropertyBase<T>& operator-=( X other )
      {
        this->set
        ( this->get
          () - other );
        return *this;
      }

      /**
       * Arithmetic assignment operator
       *
       * This method is implemented as a template for two reasons:
       *  - So that it will work with a variety of operands
       *  - So that a stub will not be generated if the operator is not defined for T
       *
       * T * X must be well defined
       */
      template <typename X>
      PropertyBase<T>& operator*=( X other )
      {
        this->set
        ( this->get
          () * other );
        return *this;
      }

      /**
       * Arithmetic assignment operator
       *
       * This method is implemented as a template for two reasons:
       *  - So that it will work with a variety of operands
       *  - So that a stub will not be generated if the operator is not defined for T
       *
       * T / X must be well defined
       */
      template <typename X>
      PropertyBase<T>& operator/=( X other )
      {
        this->set
        ( this->get
          () / other );
        return *this;
      }

      /**
       * Arithmetic assignment operator
       *
       * This method is implemented as a template for two reasons:
       *  - So that it will work with a variety of operands
       *  - So that a stub will not be generated if the operator is not defined for T
       *
       * T % X must be well defined
       */
      template <typename X>
      PropertyBase<T>& operator%=( X other )
      {
        this->set
        ( this->get
          () % other );
        return *this;
      }

      /**
       * The get_xml function is a template method with full
       * specializations for known types.
       * The default implementation returns an XML string filled
       * with the RTTI typeinfo id. This is not a portable approach.
       * However, if you provide a specialization of get_xml()
       * for your class, it _will_ be a portable solution.
       */
      virtual Glib::ustring xml( int indent = -1 )
      {
        return ReadOnlyPropertyBase<T>::xml( indent );
      }

      /**
       * Returns the RTTI type id of the contained type
       */
      const std::type_info& type() const
      {
        return ReadOnlyPropertyBase<T>::type();
      }

  };

}

#include <idioskopos/property.h>
#include <idioskopos/referenceproperty.h>

namespace Idioskopos {

/**
 * A template specialization of PropertyBase for std::vectors
 */
template <typename X, typename A>
  class PropertyBase<std::vector<X, A> >: public ReadOnlyPropertyBase<std::vector<X, A> >, public WriteOnlyPropertyBase<std::vector<X, A> >
  {
    public:
      typedef IdioskoposPointer<PropertyBase> pointer;
      typedef std::vector<X, A> value_type;

      IDIOSKOPOS_OBJECT( PropertyBase );

      /**
         * Default constructor that doesn't do much.
       */
      PropertyBase ( const Glib::ustring& name ) :
          Object( name ),
          ReadOnlyPropertyBase<std::vector<X, A> >( name ),
          WriteOnlyPropertyBase<std::vector<X, A> >( name )
      { }

      /**
       * Makes this a property of the introspectable container accessible through
       * the provided name.
       */
      PropertyBase( const Glib::ustring& name, Introspectable& container ) :
          Object( name, container ),
          ReadOnlyPropertyBase<std::vector<X, A> >( name, container ),
          WriteOnlyPropertyBase<std::vector<X, A> >( name, container )
      { }

      virtual ~PropertyBase()
      { }

      /**
       * Allows assignment to the property from the contained value type.
       */
      PropertyBase<std::vector<X, A> >& operator= ( const std::vector<X, A> & val )
      {
        this->set( val );
        return *this;
      }

      /**
       * Function call syntax
       *
       * Set the value of the property with property(value) notation.
       * @return The new value of the property
       */
      PropertyBase<std::vector<X, A> >& operator() ( const std::vector<X, A> & val )
      {
        this->set( val );
        return *this;
      }

      /**
       * The get_xml function is a template method with full
       * specializations for known types.
       * The default implementation returns an XML string filled
       * with the RTTI typeinfo id. This is not a portable approach.
       * However, if you provide a specialization of get_xml()
       * for your class, it _will_ be a portable solution.
       */
      virtual Glib::ustring xml( int indent = -1 )
      {
        return ReadOnlyPropertyBase<std::vector<X, A> >::xml( indent );
      }

      /**
       * Returns the RTTI type id of the contained type
       */
      const std::type_info& type() const
      {
        return ReadOnlyPropertyBase<std::vector<X, A> >::type();
      }

      virtual size_t num_properties() const { return this->get_pointer()->size(); }

      virtual Object::iterator properties_begin()
      {
        IdioskoposPointer< ReferenceProperty<X> > p;
        p = ReferenceProperty<X>::create(this->name()+"[0]", (*(this->get_pointer()))[0]);
        return Object::iterator(this, p, 0);
      }

    protected:
      virtual Object::iterator next_property(size_t i) {
        std::ostringstream sout;
        IdioskoposPointer< ReferenceProperty<X> > p;
        if ( i + 1 < this->get_pointer()->size() ) {
          sout << this->name() << "[" << i+1 << "]";
          p = ReferenceProperty<X>::create(sout.str(), (*(this->get_pointer()))[i+1]);
          return Object::iterator(this, p, i+1);
        }
        else
          return this->properties_end();
      }

      virtual Object::iterator previous_property(size_t i) {
        std::ostringstream sout;
        IdioskoposPointer< ReferenceProperty<X> > p;
        if ( i > 0 ) {
          sout << this->name() << "[" << i-1 << "]";
          p = ReferenceProperty<X>::create(sout.str(), (*(this->get_pointer()))[i-1]);
          return Object::iterator(this, p, i-1);
        }
        else
          return this->properties_end();
      }

  };

  // TODO Fix broken std::vector<bool> properties
  /**
   * A template specialization of PropertyBase for std::vector<bool>
   */
  template <typename A>
      class PropertyBase<std::vector<bool, A> >: public ReadOnlyPropertyBase<std::vector<bool, A> >, public WriteOnlyPropertyBase<std::vector<bool, A> >
      {
        public:
          typedef IdioskoposPointer<PropertyBase> pointer;
          typedef std::vector<bool, A> value_type;

          IDIOSKOPOS_OBJECT( PropertyBase );

      /**
           * Default constructor that doesn't do much.
       */
      PropertyBase ( const Glib::ustring& name ) :
          Object( name ),
      ReadOnlyPropertyBase<std::vector<bool, A> >( name ),
      WriteOnlyPropertyBase<std::vector<bool, A> >( name )
      { }

      /**
       * Makes this a property of the introspectable container accessible through
       * the provided name.
       */
      PropertyBase( const Glib::ustring& name, Introspectable& container ) :
          Object( name, container ),
      ReadOnlyPropertyBase<std::vector<bool, A> >( name, container ),
      WriteOnlyPropertyBase<std::vector<bool, A> >( name, container )
      { }

      virtual ~PropertyBase()
      { }

      /**
       * Allows assignment to the property from the contained value type.
       */
      PropertyBase<std::vector<bool, A> >& operator= ( const std::vector<bool, A> & val )
      {
        this->set( val );
        return *this;
      }

      /**
       * Function call syntax
       *
       * Set the value of the property with property(value) notation.
       * @return The new value of the property
       */
      PropertyBase<std::vector<bool, A> >& operator() ( const std::vector<bool, A> & val )
      {
        this->set( val );
        return *this;
      }

      /**
       * The get_xml function is a template method with full
       * specializations for known types.
       * The default implementation returns an XML string filled
       * with the RTTI typeinfo id. This is not a portable approach.
       * However, if you provide a specialization of get_xml()
       * for your class, it _will_ be a portable solution.
       */
      virtual Glib::ustring xml( int indent = -1 )
      {
        return ReadOnlyPropertyBase<std::vector<bool, A> >::xml( indent );
      }

      /**
       * Returns the RTTI type id of the contained type
       */
      const std::type_info& type() const
      {
        return ReadOnlyPropertyBase<std::vector<bool, A> >::type();
      }

//       virtual size_t num_properties() const { return this->get_pointer()->size(); }
//
//       virtual Object::iterator properties_begin()
//       {
//         IdioskoposPointer< ReferenceProperty<std::_Bit_reference> > p;
//         p = ReferenceProperty<std::_Bit_reference>::create(this->name()+"[0]", (*(this->get_pointer()))[0]);
//         return Object::iterator(this, p, 0);
//       }

//         protected:
//           virtual Object::iterator next_property(size_t i) {
//             std::ostringstream sout;
//             IdioskoposPointer< ReferenceProperty<std::_Bit_reference> > p;
//             if ( i + 1 < this->get_pointer()->size() ) {
//               sout << this->name() << "[" << i+1 << "]";
//               p = ReferenceProperty<std::_Bit_reference>::create(sout.str(), (*(this->get_pointer()))[i+1]);
//               return Object::iterator(this, p, i+1);
//             }
//             else
//               return this->properties_end();
//           }
//
//           virtual Object::iterator previous_property(size_t i) {
//             std::ostringstream sout;
//             IdioskoposPointer< ReferenceProperty<std::_Bit_reference> > p;
//             if ( i > 0 ) {
//               sout << this->name() << "[" << i-1 << "]";
//               p = ReferenceProperty<std::_Bit_reference>::create(sout.str(), (*(this->get_pointer()))[i-1]);
//               return Object::iterator(this, p, i-1);
//             }
//             else
//               return this->properties_end();
//           }

      };

   /**
    * A template specialization of PropertyBase for std::array
    */
  template <typename X, size_t N>
  class PropertyBase<IdioskoposArray<X, N> >: public ReadOnlyPropertyBase<IdioskoposArray<X, N> >, public WriteOnlyPropertyBase<IdioskoposArray<X, N> >
  {
    public:
      typedef IdioskoposPointer<PropertyBase> pointer;
      typedef IdioskoposArray<X, N> value_type;

      IDIOSKOPOS_OBJECT( PropertyBase );

      /**
             * Default constructor that doesn't do much.
       */
      PropertyBase ( const Glib::ustring& name ) :
          Object( name ),
          ReadOnlyPropertyBase<IdioskoposArray<X, N> >( name ),
          WriteOnlyPropertyBase<IdioskoposArray<X, N> >( name )
      { }

      /**
       * Makes this a property of the introspectable container accessible through
       * the provided name.
       */
      PropertyBase( const Glib::ustring& name, Introspectable& container ) :
          Object( name, container ),
          ReadOnlyPropertyBase<IdioskoposArray<X, N> >( name, container ),
          WriteOnlyPropertyBase<IdioskoposArray<X, N> >( name, container )
      { }

      virtual ~PropertyBase()
      { }

      /**
       * Allows assignment to the property from the contained value type.
       */
      PropertyBase<IdioskoposArray<X, N> >& operator= ( const IdioskoposArray<X, N> & val )
      {
        this->set( val );
        return *this;
      }

      /**
       * Function call syntax
       *
       * Set the value of the property with property(value) notation.
       * @return The new value of the property
       */
      PropertyBase<IdioskoposArray<X, N> >& operator() ( const IdioskoposArray<X, N> & val )
      {
        this->set( val );
        return *this;
      }

      /**
       * The get_xml function is a template method with full
       * specializations for known types.
       * The default implementation returns an XML string filled
       * with the RTTI typeinfo id. This is not a portable approach.
       * However, if you provide a specialization of get_xml()
       * for your class, it _will_ be a portable solution.
       */
      virtual Glib::ustring xml( int indent = -1 )
      {
        return ReadOnlyPropertyBase<IdioskoposArray<X, N> >::xml( indent );
      }

      /**
       * Returns the RTTI type id of the contained type
       */
      const std::type_info& type() const
      {
        return ReadOnlyPropertyBase<IdioskoposArray<X, N> >::type();
      }

      virtual size_t num_properties() const { return N; }

      virtual Object::iterator properties_begin()
      {
        IdioskoposPointer< ReferenceProperty<X> > p;
        p = ReferenceProperty<X>::create(this->name()+"[0]", (*(this->get_pointer()))[0]);
        return Object::iterator(this, p, 0);
      }

    protected:
      virtual Object::iterator next_property(size_t i) {
        std::ostringstream sout;
        IdioskoposPointer< ReferenceProperty<X> > p;
        if ( i + 1 < N ) {
          sout << this->name() << "[" << i+1 << "]";
          p = ReferenceProperty<X>::create(sout.str(), (*(this->get_pointer()))[i+1]);
          return Object::iterator(this, p, i+1);
        }
        else
          return this->properties_end();
      }

      virtual Object::iterator previous_property(size_t i) {
        std::ostringstream sout;
        IdioskoposPointer< ReferenceProperty<X> > p;
        if ( i > 0 ) {
          sout << this->name() << "[" << i-1 << "]";
          p = ReferenceProperty<X>::create(sout.str(), (*(this->get_pointer()))[i-1]);
          return Object::iterator(this, p, i-1);
        }
        else
          return this->properties_end();
      }

  };


}

#endif

