/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/

#ifndef IDIOSKOPOSREADONLYPROPERTYBASE_H
#define IDIOSKOPOSREADONLYPROPERTYBASE_H

#include <idioskopos/utility_stream.h>

namespace Idioskopos
{

  /**
   * Implements a simple read only property. Using this class ensures
   * that the value may not be changed through introspection methods.
   *
   * This is a pure virtual class. Children should override the get()
   * method to return the real value and the get_pointer() to return
   * a pointer to the real value.
   *
   * @author Rick L Vinyard Jr
   */
  template <typename T>
  class ReadOnlyPropertyBase: public virtual Object
  {
    public:
      typedef IdioskoposPointer<ReadOnlyPropertyBase> pointer;
      typedef T value_type;

      IDIOSKOPOS_OBJECT( ReadOnlyPropertyBase );

      /** Default constructor that doesn't do much. */
      ReadOnlyPropertyBase ( const Glib::ustring& name,
                             sigc::slot<void> slot=sigc::slot<void>() ):
          Object( name, slot )
      { }

      /**
       * Makes this a property of the introspectable container accessible through
       * the provided name.
       */
      ReadOnlyPropertyBase( const Glib::ustring& name,
                            Introspectable& container,
                            sigc::slot<void> slot=sigc::slot<void>() ):
          Object( name, container, slot )
      { }

      /** Destructor */
      virtual ~ReadOnlyPropertyBase()
      { }

      /** The canonical get() accessor method. */
      virtual const T& get() const = 0;

      /** Should return a pointer to the contained object, or NULL if pointers are not allowed */
      virtual T* get_pointer() = 0;

      virtual const T* get_pointer() const { return const_cast<ReadOnlyPropertyBase*>(this)->get_pointer(); }

      /** Allows for implicit casts to the contained type. */
      operator const T& () const { return this->get(); }

      /**
       * Function call syntax
       *
       * Get the value of the property with property() notation.
       * @return the current value of the property
       */
      const T& operator() () const { return this->get(); }

      /**
       * A convenience operator when the property contains a
       * class.
       *
       * Allows class method to be accessed using notation:
       * property->class_method()
       */
      T* operator->() { return get_pointer(); }

      /**
       * The get_xml function is a template method with full
       * specializations for known types.
       * The default implementation returns an XML string filled
       * with the RTTI typeinfo id. This is not a portable approach.
       * However, if you provide a specialization of get_xml()
       * for your class, it _will_ be a portable solution.
       */
      virtual Glib::ustring xml( int indent = -1 )
      {
        Glib::ustring s;
        for ( int i = 0; i < indent; i++ )
          s += "  ";
        s += "<property type=\"" + this->class_name() + "\" name=\"" + this->name() + "\">";
        if ( indent >= 0 )
          s += "\n";
        s += get_xml( static_cast<T>( *this ), ( ( indent >= 0 ) ? indent + 1 : indent ) );
        if ( indent >= 0 ) s += "\n";
        for ( int i = 0;
              i < indent;
              i++ ) s += "  ";
        s += "</property>";
        if ( indent >= 0 ) s += "\n";
        return s;
      }

      /**
       * Similar in concept to the xml() method, this provides a
       * string representation of the contained value if the get_value()
       * template has a meaningful template specialization.
       */
      virtual Glib::ustring value()
      {
        return get_value( static_cast<T>( *this ) );
      }

      /**
       * Returns the RTTI type id of the contained type
       */
      const std::type_info& type() const
      {
        return typeid( T );
      }

  };

  /**
   * Template specialization of ReadOnlyPropertyBase for std::vector
   */
  template <typename X, typename A>
  class ReadOnlyPropertyBase<std::vector<X, A> >: public virtual Object
  {
    public:
      typedef IdioskoposPointer<ReadOnlyPropertyBase> pointer;
      typedef std::vector<X, A> value_type;

      IDIOSKOPOS_OBJECT( ReadOnlyPropertyBase );

      /** Default constructor that doesn't do much. */
      ReadOnlyPropertyBase ( const Glib::ustring& name,
                             sigc::slot<void> slot=sigc::slot<void>() ):
          Object( name, slot )
      { }

      /**
       * Makes this a property of the introspectable container accessible through
       * the provided name.
       */
      ReadOnlyPropertyBase( const Glib::ustring& name,
                            Introspectable& container,
                            sigc::slot<void> slot=sigc::slot<void>() ):
          Object( name, container, slot )
      { }

      /** Destructor */
      virtual ~ReadOnlyPropertyBase()
      { }

      /** The canonical get() accessor method. */
      virtual const std::vector<X, A> & get() const = 0;

      /** Should return a pointer to the contained object, or NULL if pointers are not allowed */
      virtual std::vector<X, A> * get_pointer() = 0;

      virtual const std::vector<X,A>* get_pointer() const { return const_cast<ReadOnlyPropertyBase*>(this)->get_pointer(); }

      /** Allows for implicit casts to the contained type. */
      operator const std::vector<X, A> & () const { return this->get(); }

      /**
       * Function call syntax
       *
       * Get the value of the property with property() notation.
       * @return the current value of the property
       */
      const std::vector<X, A> & operator() () const { return this->get(); }

      /**
       * A convenience operator when the property contains a
       * class.
       *
       * Allows class method to be accessed using notation:
       * property->class_method()
       */
      std::vector<X, A> * operator->() { return get_pointer(); }

      /**
       * The get_xml function is a template method with full
       * specializations for known types.
       * The default implementation returns an XML string filled
       * with the RTTI typeinfo id. This is not a portable approach.
       * However, if you provide a specialization of get_xml()
       * for your class, it _will_ be a portable solution.
       */
      virtual Glib::ustring xml( int indent = -1 )
      {
        Glib::ustring s;
        for ( int i = 0; i < indent; i++ )
          s += "  ";
        s += "<property type=\"" + this->class_name() + "\" name=\"" + this->name() + "\">";
        if ( indent >= 0 )
          s += "\n";
        s += get_xml_vector( this->get
                             (), ( ( indent >= 0 ) ? indent + 1 : indent ) );
        if ( indent >= 0 )
          s += "\n";
        for ( int i = 0; i < indent; i++ )
          s += "  ";
        s += "</property>";
        if ( indent >= 0 )
          s += "\n";
        return s;
      }

      /**
       * Similar in concept to the xml() method, this provides a
       * string representation of the contained value if the get_value()
       * template has a meaningful template specialization.
       */
      virtual Glib::ustring value() { return get_value( static_cast<std::vector<X, A> >( *this ) ); }

      /**
       * Returns the RTTI type id of the contained type
       */
      const std::type_info& type() const { return typeid( std::vector<X, A> ); }

  };

  /**
   * Template specialization of ReadOnlyPropertyBase for std::array
   */
  template <typename X, size_t N>
  class ReadOnlyPropertyBase<IdioskoposArray<X,N> >: public virtual Object
  {
    public:
      typedef IdioskoposPointer<ReadOnlyPropertyBase> pointer;
      typedef IdioskoposArray<X,N> value_type;

      IDIOSKOPOS_OBJECT( ReadOnlyPropertyBase );

      /** Default constructor that doesn't do much. */
      ReadOnlyPropertyBase ( const Glib::ustring& name,
                             sigc::slot<void> slot=sigc::slot<void>() ):
          Object( name, slot )
      { }

      /**
       * Makes this a property of the introspectable container accessible through
       * the provided name.
       */
      ReadOnlyPropertyBase( const Glib::ustring& name,
                            Introspectable& container,
                            sigc::slot<void> slot=sigc::slot<void>() ):
          Object( name, container, slot )
      { }

      /** Destructor */
      virtual ~ReadOnlyPropertyBase()
      { }

      /** The canonical get() accessor method. */
      virtual const IdioskoposArray<X,N> & get() const = 0;

      /** Should return a pointer to the contained object, or NULL if pointers are not allowed */
      virtual IdioskoposArray<X,N> * get_pointer() = 0;

      virtual const IdioskoposArray<X,N>* get_pointer() const { return const_cast<ReadOnlyPropertyBase*>(this)->get_pointer(); }

      /** Allows for implicit casts to the contained type. */
      operator const IdioskoposArray<X,N> & () const
      {
        return this->get();
      }

      /**
       * Function call syntax
       *
       * Get the value of the property with property() notation.
       * @return the current value of the property
       */
      const IdioskoposArray<X,N> & operator() () const
      {
        return this->get
        ();
      }

      /**
       * A convenience operator when the property contains a
       * class.
       *
       * Allows class method to be accessed using notation:
       * property->class_method()
       */
      IdioskoposArray<X,N> * operator->()
      {
        return get_pointer();
      }

      /**
       * The get_xml function is a template method with full
       * specializations for known types.
       * The default implementation returns an XML string filled
       * with the RTTI typeinfo id. This is not a portable approach.
       * However, if you provide a specialization of get_xml()
       * for your class, it _will_ be a portable solution.
       */
      virtual Glib::ustring xml( int indent = -1 )
      {
        Glib::ustring s;
        for ( int i = 0; i < indent; i++ )
          s += "  ";
        s += "<property type=\"" + this->class_name() + "\" name=\"" + this->name() + "\">";
        if ( indent >= 0 )
          s += "\n";
        s += get_xml( this->get(), ( ( indent >= 0 ) ? indent + 1 : indent ) );
        if ( indent >= 0 )
          s += "\n";
        for ( int i = 0; i < indent; i++ )
          s += "  ";
        s += "</property>";
        if ( indent >= 0 )
          s += "\n";
        return s;
      }

      /**
       * Similar in concept to the xml() method, this provides a
       * string representation of the contained value if the get_value()
       * template has a meaningful template specialization.
       */
      virtual Glib::ustring value()
      {
        return get_value( static_cast<IdioskoposArray<X,N> >( *this ) );
      }

      /**
       * Returns the RTTI type id of the contained type
       */
      const std::type_info& type() const
      {
        return typeid( IdioskoposArray<X,N> );
      }

  };

}

#endif

