/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/

#ifndef IDIOSKOPOSREFERENCEPROPERTY_H
#define IDIOSKOPOSREFERENCEPROPERTY_H

#include <idioskopos/readonlypropertybase.h>
#include <idioskopos/writeonlypropertybase.h>
#include <idioskopos/propertybase.h>

namespace Idioskopos {

  /**
   * Read only reference properties
   */
  template <typename T>
  class ReadOnlyReferenceProperty: public ReadOnlyPropertyBase<T> {
    public:
      typedef IdioskoposPointer<ReadOnlyReferenceProperty> pointer;
      typedef T value_type;

      IDIOSKOPOS_OBJECT(ReadOnlyReferenceProperty);

      /**
       * Initializes the contained value to the parameter's value.
       */
      ReadOnlyReferenceProperty ( const Glib::ustring& name,
                                  T& val,
                                  sigc::slot<void> slot=sigc::slot<void>() ) :
          Object(name, slot),
          ReadOnlyPropertyBase<T>(name),
          m_value( val ) { }

      /**
       * Makes this a property of the introspectable container accessible through
       * the provided name, and initializes the contained value to the value of
       * the third parameter.
       */
      ReadOnlyReferenceProperty ( const Glib::ustring& name,
                                  T& val,
                                  Introspectable& container,
                                  sigc::slot<void> slot=sigc::slot<void>() ) :
          Object(name, container, slot),
          ReadOnlyPropertyBase<T>( name, container ),
          m_value( val )
      { }

      static pointer create( const Glib::ustring& name,
                             T& val,
                             sigc::slot<void> slot=sigc::slot<void>() )
      {
        return pointer( new ReadOnlyReferenceProperty<T>(name, val, slot) );
      }

      static pointer create( const Glib::ustring& name,
                             T& val,
                             Introspectable& container,
                             sigc::slot<void> slot=sigc::slot<void>() )
      {
        return pointer( new ReadOnlyReferenceProperty<T>(name, val, container, slot) );
      }

      virtual ~ReadOnlyReferenceProperty() { }

      /** The canonical get() accessor method. */
      virtual const T& get() const { return m_value; }

      /** Should return a pointer to the contained object, or NULL if pointers are not allowed */
      virtual T* get_pointer() {
        return & m_value;
      }

    protected:
      T& m_value;

  };

  /**
   * Write only reference properties
   */
  template <typename T>
  class WriteOnlyReferenceProperty: public WriteOnlyPropertyBase<T> {
    public:
      typedef IdioskoposPointer<WriteOnlyReferenceProperty> pointer;
      typedef T value_type;

      IDIOSKOPOS_OBJECT(WriteOnlyReferenceProperty);

      /**
       * Initializes the contained value to the parameter's value.
       */
      WriteOnlyReferenceProperty ( const Glib::ustring& name,
                                   T& val,
                                   sigc::slot<void> slot=sigc::slot<void>() ) :
          Object(name, slot),
          WriteOnlyPropertyBase<T>(name,val),
          m_value( val ) { }

      /**
       * Makes this a property of the introspectable container accessible through
       * the provided name, and initializes the contained value to the value of
       * the third parameter.
       */
      WriteOnlyReferenceProperty ( const Glib::ustring& name,
                                   T& val,
                                   Introspectable& container,
                                   sigc::slot<void> slot=sigc::slot<void>() ) :
              Object(name, container, slot),
              WriteOnlyPropertyBase<T>( name, container ),
              m_value( val ) { }

      static pointer create( const Glib::ustring& name,
                             T& val,
                             sigc::slot<void> slot=sigc::slot<void>() )
      {
        return pointer( new WriteOnlyReferenceProperty<T>(name, val, slot) );
      }

      static pointer create( const Glib::ustring& name,
                             T& val,
                             Introspectable& container,
                             sigc::slot<void> slot=sigc::slot<void>() )
      {
        return pointer( new WriteOnlyReferenceProperty<T>(name, val, container, slot) );
      }

      virtual ~WriteOnlyReferenceProperty() { }

      /** The canonical set() accessor method that also emits the changed signal. */
      virtual void set
        ( const T& val ) {
        m_value = val;
        this->signal_value_changed().emit();
      }

    protected:
      T& m_value;

      virtual T& get_value() { return m_value; }
  };

  /**
   * Reference properties
   */
  template <typename T>
  class ReferenceProperty: public PropertyBase<T> {
    public:
      typedef IdioskoposPointer<ReferenceProperty> pointer;
      typedef T value_type;

      IDIOSKOPOS_OBJECT(ReferenceProperty);

      /**
       * Initializes the contained value to the parameter's value.
       */
      ReferenceProperty ( const Glib::ustring& name,
                          T& val,
                          sigc::slot<void> slot=sigc::slot<void>() ) :
          Object(name, slot),
          PropertyBase<T>(name),
          m_value( val ) { }

      /**
       * Makes this a property of the introspectable container accessible through
       * the provided name, and initializes the contained value to the value of
       * the third parameter.
       */
      ReferenceProperty ( const Glib::ustring& name,
                          T& val,
                          Introspectable& container,
                          sigc::slot<void> slot=sigc::slot<void>() ) :
          Object( name, container, slot ),
          PropertyBase<T>(name, container),
          m_value( val ) { }

      static pointer create( const Glib::ustring& name,
                             T& val,
                             sigc::slot<void> slot=sigc::slot<void>() )
      {
        return pointer( new ReferenceProperty<T>(name, val, slot) );
      }

      static pointer create( const Glib::ustring& name,
                             T& val,
                             Introspectable& container,
                             sigc::slot<void> slot=sigc::slot<void>() )
      {
        return pointer( new ReferenceProperty<T>(name, val, container, slot) );
      }

      virtual ~ReferenceProperty() { }

      /** The canonical get() accessor method. */
      virtual const T& get() const { return m_value; }

      /** Should return a pointer to the contained object, or NULL if pointers are not allowed */
      virtual T* get_pointer() {
        return & m_value;
      }

      /** The canonical set() accessor method that also emits the changed signal. */
      virtual void set( const T& val ) {
        m_value = val;
        this->signal_value_changed().emit();
      }

    protected:
      T& m_value;

      virtual T& get_value() { return m_value; }
  };

}


#endif

