/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#ifndef IDIOSKOPOSUTILITY_STREAM_H
#define IDIOSKOPOSUTILITY_STREAM_H

#include <sstream>
#include <vector>
#include <idioskopos/array.h>

namespace Idioskopos {

  // TODO Switch to a name registration rather than these overloaded functions

/**
 * This template function returns an xml string of the property
 * and the contained value. Since the default template returns
 * a value for type that is filled with RTTI information this
 * is definiately not a portable approach. However, if a full
 * specialization is provided, it becomes a portable solution
 * for those types that have template specializations.
 *
 * The macro IDIOSKOPOS_XMLSTRING is provided to make creating
 * template specializations easier. The first parameter to the
 * macro is your type, and the second parameter is the typeid
 * string.
 *
 * The macro IDIOSKOPOS_STREAMABLE creates template specializations
 * for both the get_xmlstring() and get_valuestring() template functions.
 */
template <typename T>
  inline Glib::ustring get_xml(const T& t, int indent=-1) {
    std::ostringstream sout;
    for (int i=0; i < indent; i++ ) sout << "  ";
    sout << "<value type=\""
         << typeid(T).name() << "\"/>";
    return sout.str();
  }

template <typename T,typename A>
  inline Glib::ustring get_xml_vector( const std::vector<T,A>& a, int indent=-1 ) {
    std::ostringstream sout;
    int i;
    size_t n;
    for (i=0; i < indent; i++ ) sout << "  ";
    sout << "<vector type=\""
        << typeid(T).name() << "\" size=\""
        << a.size() << "\">";
    for (n=0; n < a.size(); n++) {
      if ( indent >= 0 ) sout << "\n";
      sout << get_xml( a[n], (indent>=0)?indent+1:indent );
    }
    if (indent >= 0 ) sout << "\n";
    for (i=0; i < indent; i++ ) sout << "  ";
    sout << "</vector>";
    return sout.str();
  }

template <typename T, size_t N>
  inline Glib::ustring get_xml( const IdioskoposArray<T,N>& a, int indent=-1 ) {
    std::ostringstream sout;
    int i;
    size_t n;
    for (i=0; i < indent; i++ ) sout << "  ";
    sout << "<array type=\""
         << typeid(T).name() << "\" size=\""
         << N << "\">";
    for (n=0; n < N; n++) {
      if ( indent >= 0 ) sout << "\n";
      sout << get_xml( a[n], (indent>=0)?indent+1:indent );
    }
    if (indent >= 0 ) sout << "\n";
    for (i=0; i < indent; i++ ) sout << "  ";
    sout << "</array>";
    return sout.str();
  }


/**
 * A convenience macro for that defines a template specialization for
 * the get_xmlstring() function template, associating a given string
 * with a type.
 */
#define IDIOSKOPOS_XML( Type, TypeString )                                           \
inline Glib::ustring get_xml(const Type& t, int indent=-1) {                         \
  std::ostringstream out;                                                            \
  for (int i=0; i < indent; i++ )                                                    \
      out << "  ";                                                                   \
  out << "<value type=\""                                                            \
      << TypeString << "\">"                                                         \
      << t << "</value>";                                                            \
  return out.str();                                                                  \
  }                                                                                  \
                                                                                     \
template <size_t N>                                                                  \
  inline Glib::ustring get_xml( const IdioskoposArray<Type,N>& a, int indent=-1 ) {  \
    std::ostringstream sout;                                                         \
    int i;                                                                           \
    size_t n;                                                                        \
    for (i=0; i < indent; i++ ) sout << "  ";                                        \
    sout << "<array type=\""                                                         \
         << TypeString << "\" size=\""                                               \
         << N << "\">";                                                              \
    for (n=0; n < N; n++) {                                                          \
      if ( indent >= 0 ) sout << "\n";                                               \
      sout << get_xml( a[n], (indent>=0)?indent+1:indent );                          \
    }                                                                                \
    if (indent >= 0 ) sout << "\n";                                                  \
    for (i=0; i < indent; i++ ) sout << "  ";                                        \
    sout << "</array>";                                                              \
    return sout.str();                                                               \
  }                                                                                  \
                                                                                     \
template <typename A>                                                                \
inline Glib::ustring get_xml_vector( const std::vector<Type,A>& a, int indent=-1 ) { \
  std::ostringstream sout;                                                           \
  int i;                                                                             \
  size_t n;                                                                          \
  for (i=0; i < indent; i++ ) sout << "  ";                                          \
  sout << "<vector type=\""                                                          \
       << TypeString << "\" size=\""                                                 \
       << a.size() << "\">";                                                         \
  for (n = 0; n < a.size(); n++) {                                                   \
    if ( indent >= 0 ) sout << "\n";                                                 \
    sout << get_xml( a[n], (indent>=0)?indent+1:indent );                            \
  }                                                                                  \
  if (indent >= 0 ) sout << "\n";                                                    \
  for (i=0; i < indent; i++ ) sout << "  ";                                          \
  sout << "</vector>";                                                               \
  return sout.str();                                                                 \
 }


template<typename T>
  inline Glib::ustring get_value(const T& t) {
    return Glib::ustring();
  }

/**
 * A convenience macro for that defines a template specialization for
 * the get_value() function template.
 */
#define IDIOSKOPOS_GET_VALUE( Type )             \
inline Glib::ustring get_value(const Type& t) {  \
  std::ostringstream sout;                       \
  sout << t;                                     \
  return sout.str();                             \
  }


template<typename T>
inline void set_value(const Glib::ustring& s, T& t) {
  return;
}

/**
 * A convenience macro for that defines a template specialization for
 * the set_value() function template.
 */
#define IDIOSKOPOS_SET_VALUE( Type )                        \
  inline void set_value(const Glib::ustring& s, Type& t) {  \
  std::istringstream sin(s);                                \
  sin >> t;                                                 \
  return;                                                   \
}

/**
 * A convenience macro for that defines template specializations for
 * both get_xmlstring() and get_valuestring() function templates.
 */
#define IDIOSKOPOS_STREAMABLE( Type, TypeString )  \
  IDIOSKOPOS_XML( Type, TypeString )               \
  IDIOSKOPOS_GET_VALUE(Type )                      \
  IDIOSKOPOS_SET_VALUE(Type )

IDIOSKOPOS_STREAMABLE(float,"float");
IDIOSKOPOS_STREAMABLE(double,"double");
IDIOSKOPOS_STREAMABLE(long double,"long double");
IDIOSKOPOS_STREAMABLE(short int,"short int");
IDIOSKOPOS_STREAMABLE(unsigned short int,"unsigned short int");
IDIOSKOPOS_STREAMABLE(int,"int");
IDIOSKOPOS_STREAMABLE(unsigned int,"unsigned int");
IDIOSKOPOS_STREAMABLE(long int,"long int");
IDIOSKOPOS_STREAMABLE(unsigned long int,"unsigned long int");
IDIOSKOPOS_STREAMABLE(char,"char");
IDIOSKOPOS_STREAMABLE(unsigned char,"unsigned char");
IDIOSKOPOS_STREAMABLE(bool,"bool");
IDIOSKOPOS_STREAMABLE(std::string,"std::string");
IDIOSKOPOS_STREAMABLE(Glib::ustring,"Glib::ustring");

}

#endif

